# Simulation Aids project demonstration code
#
# Copyright 2015 Automation Professionals, LLC <sales@automation-pros.com>
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   2. Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#   3. The name of the author may not be used to endorse or promote products
#      derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.

stepmap = {}

class StepStatus:
	"""Track current position and timing in a dataset of steps of value/duration"""
	def __init__(self, hash):
		from java.util import Date
		self.step = 0
		self.time = Date().time
		stepmap[hash] = self

	def fixStep(self, dataset, ts):
		"""Reset the current step number if it exceeds the highest row index
		in the given dataset.  Returns True if there are no rows."""
		if self.step >= dataset.rowCount:
			self.step = 0
			self.time = ts
		return self.step >= dataset.rowCount

	def evaluate(self, binding, dataset):
		"""Determine the current step, schedule an update when it expires, and
		return the current value.  Returns zero if no step defined."""
		from java.util import Date
		ts = Date().time
		if self.fixStep(dataset, ts):
			return 0.0
		durations = [int(dataset.getValueAt(x, 'duration')) for x in range(dataset.rowCount)]
		endts = durations[self.step] + self.time
		if ts>endts:
			cycle = sum(durations)
			if cycle<=0:
				raise RuntimeError("Total duration of all steps must be positive")
			while ts>endts:
				self.time = endts
				self.step = (self.step + 1) % len(durations)
				endts = durations[self.step] + self.time
		system.util.invokeLater(binding.childInteractionUpdated, endts-ts)
		return dataset.getValueAt(self.step, 'value')

def evaluateStep(binding, dataset):
	hash = id(binding)
	print "Binding Hash: "+str(hash)
	if hash in stepmap:
		return stepmap[hash].evaluate(binding, dataset)
	return StepStatus(hash).evaluate(binding, dataset)

def stepEditPopup(event):
	if not event.popupTrigger:
		return
	table = event.source
	menuNames = []
	menuFunctions = []
	if table.selectedRow>=0 and table.selectedRow<table.data.rowCount:
		def deleteRow(event, comp=table):
			comp.data = system.dataset.deleteRow(comp.data, comp.selectedRow)
		menuNames.append('Delete Row')
		menuFunctions.append(deleteRow)
		def dupeRow(event, comp=table):
			data = comp.data
			row = [data.getValueAt(table.selectedRow, x) for x in range(data.columnCount)]
			comp.data = system.dataset.addRow(data, table.selectedRow, row)
		menuNames.append('Duplicate Row')
		menuFunctions.append(dupeRow)
	def addRow(event, comp=table):
		data = comp.data
		row = ['New Label', 0.0, 5000]
		comp.data = system.dataset.addRow(data, row)
	menuNames.append('Add Row')
	menuFunctions.append(addRow)
	popupMenu = system.gui.createPopupMenu(menuNames, menuFunctions)
	popupMenu.show(event)

def coalesce(*args):
	for x in args:
		if x is not None:
			return x
	return None

animateMap = {}

def animateShape(binding, x=None, y=None, w=None, h=None, pace=5.0, ms=50):
	shape = binding.target
	x0, y0, xx0, yy0 = (shape.relX, shape.relY, shape.relX+shape.relWidth, shape.relY+shape.relHeight)
	x, y, xx, yy = (coalesce(x, x0), coalesce(y, y0), coalesce(x+w, yy0), coalesce(y+h, yy0))
	deltas = (x-x0, y-y0, xx-xx0, yy-yy0)
	dmax = max([abs(a) for a in deltas])
	if dmax>pace:
		rpace = pace/dmax
		sx, sy, sxx, syy = tuple([a+b*rpace for a,b in zip((x0, y0, xx0, yy0), deltas)])
		system.util.invokeLater(binding.childInteractionUpdated, ms)
		shape.relX, shape.relY, shape.relWidth, shape.relHeight = (sx, sy, sxx-sx, syy-sy)
		return ms/rpace - ms
	shape.relX, shape.relY, shape.relWidth, shape.relHeight = (x, y, w, h)
	return 0

def animateComponent(binding, x=None, y=None, w=None, h=None, pace=5.0, ms=50):
	comp = binding.target
	hash = id(binding)
	p0 = animateMap.get(hash)
	if p0 is None:
		p0 = {}
	else:
		del animateMap[hash]
	x0, y0, w0, h0 = (p0.get('x', float(comp.x)), p0.get('y', float(comp.y)), p0.get('w', float(comp.width)), p0.get('h', float(comp.height)))
	x, y, w, h = (float(coalesce(x, x0)), float(coalesce(y, y0)), float(coalesce(w, y0)), float(coalesce(h, y0)))
	deltas = (x-x0, y-y0, w-w0, h-h0)
	dmax = max([abs(a) for a in deltas])
	if dmax>pace:
		rpace = pace/dmax
		sx, sy, sw, sh = tuple([a+b*rpace for a,b in zip((x0, y0, w0, h0), deltas)])
		animateMap[hash] = {'x': sx, 'y': sy, 'w': sw, 'h': sh}
		system.util.invokeLater(binding.childInteractionUpdated, ms)
		system.gui.reshapeComponent(comp, int(sx), int(sy), int(sw), int(sh))
		return ms/rpace - ms
	system.gui.reshapeComponent(comp, int(x), int(y), int(w), int(h)) 
	return 0
